# Plugin Artifact Resolution

**Owners:**
Pepper Lebeck-Jobe [@eljobe](https://github.com/eljobe)
Stefan Oehme [@oehme](https://github.com/oehme)   
**Updated:** 2016-05-23   
![Approved](https://img.shields.io/badge/design-approved-green.svg)

<!-- START doctoc generated TOC please keep comment here to allow auto update -->
<!-- DON'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->


- [Objective](#objective)
  - [Goals / Use Cases](#goals--use-cases)
  - [Non-Goals / Out of Scope Use Cases](#non-goals--out-of-scope-use-cases)
- [Background](#background)
- [Solution](#solution)
  - [Plugin marker artifacts in the repository](#plugin-marker-artifacts-in-the-repository)
  - [New CustomRepositoryPluginResolver in PluginResolverFactory](#new-customrepositorypluginresolver-in-pluginresolverfactory)
- [Testing Plan](#testing-plan)
- [Open Questions](#open-questions)

<!-- END doctoc generated TOC please keep comment here to allow auto update -->

## Objective

Milestone 1.1 of the [stable-plugins-dsl project](https://github.com/gradle/stable-plugins-dsl) allows users who
maintain their own artifact repositories (either Maven or Ivy) to store the artifacts which contain their proprietary
Gradle plugins in that repository (with some additional metadata) and have them resolved during a Gradle build just by
specifying them in the `plugins` block of our Domain Specific Language (DSL.)

### Goals / Use Cases

* Users who have plugin artifacts in an artifact repository with sufficient metadata to map from pluginid and version
can use the `plugins` block of the [Gradle DSL](https://docs.gradle.org/current/userguide/plugins.html#sec:plugins_block)
to specify their plugins and Gradle will automatically resolve the correct set of artifacts from the repository.

### Non-Goals / Out of Scope Use Cases

* Users can specify their private artifact repositories in the Gradle DSL.
  * This is covered by
    [Milestone 1.2](M1.2.md)
    (M1.2) which is part of this effort, but not the concern of this design. For now, we support a single
    additional repository via a Java System property, `org.gradle.plugin.repoUrl` which can be set to an absolute
    URL, or a relative path on the filesystem. This is a temporary mechanism, which will be obviated by the
    completion of M1.2.
* Plugin classpaths are isolated from each other. The initial implementation will put all plugins on the buildscript
  classpath.

## Background

With the old way of declaring plugins which were needed during a build, users had to specify the repositories to search
for plugin artifacts, explicit dependencies on those artifacts and, ultimately the name of the plugins inside the artifacts
which needed to be applied to the build.

The introduction of the `plugins` block in the DSL and the [Gradle Plugins Portal](https://plugis.gradle.org), removed the
need for specifying either the repositories or the artifacts by introducing a service which could map between pluginid and
version numbers and the associated artifacts. And this works great for plugins which are publicly registered in the Plugins
Portal. For users who have proprietary plugins living in on-premises artifact repositories, they have had to fall back to
using the buildscript’s repositories and dependencies blocks to specify the locations of their plugins.

You need to be familiar with how artifacts are stored and addressed in Maven and Ivy artifact repositories to understand
how we are planning to tackle this problem. Here are some references:

* [https://maven.apache.org/pom.html#Maven_Coordinates](https://maven.apache.org/pom.html#Maven_Coordinates)
* [http://ant.apache.org/ivy/history/latest-milestone/concept.html](http://ant.apache.org/ivy/history/latest-milestone/concept.html)

Gradle Plugins have a many-to-one relationship with repository artifacts. That is to say, you can bundle many different
Gradle Plugins into a single jar file and put that in an artifact repository.

## Solution

This design calls for two high-level technical changes to the Gradle system, which work together to achieve the goals of
this design.

1. Plugin marker artifacts in the repository
2. New CustomRepositoryPluginResolver in PluginResolverFactory

### Plugin marker artifacts in the repository

We would like to be able to look at a plugin declaration like this one:

```gradle
plugins {
  id "fliberty-jibbet" version "42.1"
}
```

and be able to stably map that onto a specific artifact in an artifact repository and its dependencies.

To accomplish this goal, we will create "plugin marker" artifacts which are artifacts in the repository which depend on the
artifact which actually holds the plugin. In the following table, you can see how we generate the identifiers for the
artifact from the declaration of the plugin.

| Plugin Datum          | Maven Concept | Ivy Concept |
| :-----------          | :------------ | :---------- |
| id                    | groupId       | org         |
| id + '.gradle.plugin' | artifactId    | name        |
| version               | version       | rev         |

So, in our example above, we would attempt to resolve a "plugin marker" in the maven repository with this format:
`fliberty-jibbet:fliberty-jibbet:42.1` which would in turn depend on the artifact which actually holds the implementation
of the plugin.

This design is not concerned with publishing these plugin artifacts. That is covered in M1.3 of this same project.

### New CustomRepositoryPluginResolver in PluginResolverFactory

The actual work of resolving plugins declared in the plugins DSL element is handled by a chain of
[PluginResolver](https://github.com/gradle/gradle/blob/master/subprojects/plugin-use/src/main/java/org/gradle/plugin/use/resolve/internal/PluginResolver.java)
implementations which look in various places one after the other attempting to resolve the plugins. The order is
configured in the implementation of
[PluginResolverFacotry.addDefaultResolvers()](https://github.com/gradle/gradle/blob/master/subprojects/plugin-use/src/main/java/org/gradle/plugin/use/internal/PluginResolverFactory.java#L57).

This design calls for introducing a new CustomRepositoryPluginResolver just before the PluginResolutionServiceResolver.
The later is the resolver responsible for finding plugins in the Gradle Plugin Portal. We should prefer using plugins found
in any custom repositories the user has specified to finding them in the Gradle Plugin Portal. There are a couple of
scenarios which inform this decision:

* Access to the custom repository is often faster than contacting the plugin portal because it is closer.
  * This also means that sometimes, customers with lots of external dependencies actually set up a local mirror of
    popular plugins.
* Users may want to make minor tweaks to plugin implementations for their internal systems but use the same artifact
  versions so they can "mask" the publicly available plugins.

The CustomRepositoryPluginResolver consults the undocumented system property org.gradle.plugin.repoUrl to find the url of
a Maven repository which contains plugins as well as plugin markers. The system property is a placeholder implementation
for answering the question "Which repositories should we look in to find plugins."
[M1.2](M1.2.md) will obviate this mechanism.
To ensure that reading the system property doesn't cause performance problems, it is read "once" in a thread-friendly
implementation and cached in a local member of the instance of the resolver.

The CustomRepositoryPluginResolver short-circuits if the system property hasn't been set so that the next plugin resolver
in the chain may attempt to resolve the plugin artifacts. Short-circuiting rather than setting the PluginResolutionResult
to notFound prevents users from seeing the this resolver listed as a place which was searched in the "plugin not found"
message resulting from a missing plugin.

There are some constraints on the plugin requests the CustomRepositoryPluginResolver can handle:

* The plugin request must have both a plugin id and a plugin version.
  [The documentation](https://docs.gradle.org/current/userguide/plugins.html#sec:plugins_block) is clear that only plugins
  distributed with Gradle are allowed to omit the version declaration.
* SNAPSHOT and dynamic versions are not supported yet, but are called for in unconfirmed milestones of the
  [overall project roadmap](https://docs.google.com/document/d/18w0SkzPXYKOvDL85ixAynmKMAasa6akb2voum56uAuw/edit#heading=h.vds5ki6jq1cl).

Once the constraints are checked, the resolver simply constructs a Maven "group:artifact:version" string and passes the
work of resolving the plugin artifact and its dependencies on to a dependency resolver which will pull the transitive
dependencies of the "plugin marker" artifact on the the classpath which will be used for loading plugins.

## Testing Plan

This section covers the test-cases we need to ensure we have captured in our unit and integration tests.

* Can resolve a plugin artifact from a maven repository which contains a plugin marker which depends on the correct
  artifact when system property is set.
* Custom repositories are considered before the plugin portal
* Plugin resolution is not attempted when system property is not set.
* If the marker cannot be resolved, it is reported as "notFound" (instead of e.g. failing the build with a dependency
  resolution exception)
* When a plugin cannot be found in the custom repository, we fall back to the Gradle Plugin Portal.
* A plugin’s API can be accessed from the build script it is applied to
* Plugin requests without versions are rejected.
* Plugin requests with SNAPSHOT versions are rejected.
* Plugin requests with dynamic versions are rejected
* A malformed repoUrl causes the build to fail

## Open Questions

1. Do we need to handle resolving from Ivy repositories for M1.1?
  * Once we have [M1.2](M1.2.md),
    we'll automatically pick up the ability to resolve from Ivy repositories, so we are considering not supporting
    it explicitly during M1.1
2. If we allow multiple repositories, should there be one "not found" message with the description “User-defined Plugin
   Repositories” or should there be one “not found” message for each repository?
  * One for each repository.
3. What do we want to do if the plugin marker points to an artifact that doesn't contain the plugin in its transitive
   dependencies? How good of an error message can we hope to present?
  * This is too rare of a corner-case to worry about.
